import { AssetManager } from "./asset-manager.js";
import { InlineStoreEntry } from "../kv-store/inline-store-entry.js";
const decoder = new TextDecoder();
export function findMatchingSourceAndInfo(acceptEncodingsGroups, defaultSourceAndInfo, sourceAndInfoForEncodingFn) {
    let sourceAndInfo;
    let contentEncoding = null;
    if (acceptEncodingsGroups != null) {
        for (const encodingGroup of acceptEncodingsGroups) {
            const sourceAndInfosForEncodingsGroup = encodingGroup
                .map(encoding => ({ encoding, sourceAndInfo: sourceAndInfoForEncodingFn(encoding) }))
                .filter(x => x.sourceAndInfo != null);
            if (sourceAndInfosForEncodingsGroup.length === 0) {
                // If no encoding in this group is available then move to next group
                continue;
            }
            // Sort the items, putting the smallest size first
            sourceAndInfosForEncodingsGroup
                .sort((a, b) => a.sourceAndInfo.size - b.sourceAndInfo.size);
            // The first item is the one we want.
            sourceAndInfo = sourceAndInfosForEncodingsGroup[0].sourceAndInfo;
            contentEncoding = sourceAndInfosForEncodingsGroup[0].encoding;
            break;
        }
    }
    sourceAndInfo ?? (sourceAndInfo = defaultSourceAndInfo);
    return { sourceAndInfo, contentEncoding };
}
class ContentRuntimeAsset {
    constructor(metadata, sourceAndInfo) {
        this.metadata = metadata;
        this.sourceAndInfo = sourceAndInfo;
    }
    get isLocal() {
        return true;
    }
    get assetKey() {
        return this.metadata.assetKey;
    }
    async getStoreEntry() {
        const { source, hash, size } = this.sourceAndInfo;
        return new InlineStoreEntry(source, null, hash, size);
    }
    getBytes() {
        return this.sourceAndInfo.source;
    }
    getText() {
        if (!this.metadata.text) {
            throw new Error("Can't getText() for non-text content");
        }
        return decoder.decode(this.sourceAndInfo.source);
    }
    getMetadata() {
        return this.metadata;
    }
}
export class ContentBytesAsset extends ContentRuntimeAsset {
    constructor(metadata) {
        super(metadata, {
            source: metadata.fileInfo.bytes,
            hash: metadata.fileInfo.hash,
            size: metadata.fileInfo.size,
        });
        this.type = 'bytes';
    }
}
class ContentStringAsset extends ContentRuntimeAsset {
    constructor(metadata) {
        super(metadata, {
            source: ContentStringAsset.encoder.encode(metadata.fileInfo.content),
            hash: metadata.fileInfo.hash,
            size: metadata.fileInfo.size,
        });
        this.type = 'string';
    }
}
ContentStringAsset.encoder = new TextEncoder();
export { ContentStringAsset };
class ContentAssets extends AssetManager {
    constructor(contentAssetMetadataMap, assetBuilderContext = {}) {
        super();
        for (const [assetKey, metadata] of Object.entries(contentAssetMetadataMap)) {
            if (!(metadata.type in ContentAssets.builders)) {
                throw new Error(`Unknown content asset type '${metadata.type}'`);
            }
            const asset = ContentAssets.builders[metadata.type](metadata, assetBuilderContext);
            this.initAsset(assetKey, asset);
        }
    }
    static registerAssetBuilder(type, builder) {
        this.builders[type] = builder;
    }
}
ContentAssets.builders = {};
export { ContentAssets };
ContentAssets.registerAssetBuilder('bytes', metadata => new ContentBytesAsset(metadata));
ContentAssets.registerAssetBuilder('string', metadata => new ContentStringAsset(metadata));
